// -------------------------------------------------------
//
// Author: Thomas Lenart    
// Email:  thomas.lenart@es.lth.se
// Date:   March 2003
//
// -------------------------------------------------------
#include "main.h"

//#define IS_HARDWARE_ACCELERATED

#include <stdio.h>

#include "memhandler.h"
#include "frames.h"
#include "constants.h"
#include "huffman.h"
#include "requantizer.h"
#include "fixpoint.h"
#include "reorder.h"
#include "stereo.h"
#include "alias.h"
#include "imdct.h"
#include "inversion.h"
#include "subbandsyn.h"

static int random_block = START_BLOCK;
static int block_type_sequence[] = {0, 0, 2, 2, 1, 2, 3, 2, 2, 3, 3, 0, 1, 2, 1, 1};

int find_frame_sync(CT_Bitstream *bitstream)
{
    int syncword, mask;
	
    mask = (1 << 12) - 1;			 // calculate mask 1111 1111 1111
    syncword = bitstream_read_16(bitstream, 12); // read a 12 bit sequence from the stream

    while (!bitstream_eos(bitstream)) {		// while not end of bitstream
	if ((syncword & mask) == mask)		// check if syncword found
	    return (bitstream_pos(bitstream) - 12); // if true, return bitstream pos for start of sync word
	syncword = (syncword << 1) + bitstream_read_1(bitstream); // move left and add the next bit
// 	printf("hahahaha: %d\n", bitstream_pos(bitstream));
    }

    return -1;		// did not find any syncword in the stream
}

int frame_header(CT_Bitstream *bitstream, CT_MPEGInfo *frame_info)
{
    frame_info->header.ID		= bitstream_read_1(bitstream);	// always 1
    frame_info->header.layer	= bitstream_read(bitstream, 2);
    frame_info->header.protection	= 1 - bitstream_read_1(bitstream);// 0 = protection, 1 = no protection
    frame_info->header.bitrate_index	= bitstream_read(bitstream, 4);
    frame_info->header.sampling_frequency	= bitstream_read(bitstream, 2);
    frame_info->header.padding_bit		= bitstream_read_1(bitstream);
    frame_info->header.private_bit	= bitstream_read_1(bitstream);
    frame_info->header.mode		= bitstream_read(bitstream, 2);
    frame_info->header.mode_extension	= bitstream_read(bitstream, 2);
    frame_info->header.copyright	= bitstream_read_1(bitstream);
    frame_info->header.original		= bitstream_read_1(bitstream);
    frame_info->header.emphasis		= bitstream_read(bitstream, 2);

/* 	printf("haha: %X, %X, %X, %X\n", frame_info->header.ID, frame_info->header.layer, frame_info->header.bitrate_index, frame_info->header.sampling_frequency); */

    if (frame_info->header.mode == 3)		// if single channel
	frame_info->num_channels = 1;		// set to mono
    else
	frame_info->num_channels = 2;		// set to stereo

    if (!frame_info->header.ID)			// reserved bit, not ISO 11172-3
	return 0;	
    if (!frame_info->header.layer)			// reserved layer value
	return 0;
    if (frame_info->header.bitrate_index == 15)	// forbidden bitrate index
	return 0;
    if (frame_info->header.sampling_frequency == 3) // forbidden sampling frequency index
	return 0;	

    return 1;				// number of bits read
}

int frame_error_check(CT_Bitstream *bitstream, CT_MPEGInfo *frame_info)
{
    if (frame_info->header.protection) {		// if error protected
	frame_info->crc = bitstream_read_16(bitstream, 16); // read the 16 bit crc code

	// todo : CRC calculation
	return 1;	// return number of bits read
    }

    return 1;		// return number of bits read
}

void frame_audio_data(CT_Bitstream *bitstream, CT_MPEGInfo *frame_info)
{
    char	ch, gr, window, region, scfsi_band, block_type;
	
    frame_info->main_data_begin = bitstream_read_16(bitstream, 9);

    if (frame_info->header.mode == 3)
	frame_info->private_bits = bitstream_read(bitstream, 5);
    else
	frame_info->private_bits = bitstream_read(bitstream, 3);
    
    for (ch = 0; ch < frame_info->num_channels; ch++) 
	for (scfsi_band = 0; scfsi_band < 4; scfsi_band++) 
	    frame_info->scfsi[ch][scfsi_band] = bitstream_read_1(bitstream);
			
    for (gr = 0; gr < 2; gr++) 
	for (ch = 0; ch < frame_info->num_channels; ch++) {
	    frame_info->gr[gr].ch[ch].part2_3_length = bitstream_read_16(bitstream, 12);
	    frame_info->gr[gr].ch[ch].big_values     = bitstream_read_16(bitstream, 9);
	    frame_info->gr[gr].ch[ch].global_gain    = bitstream_read(bitstream, 8);
	    frame_info->gr[gr].ch[ch].scalefac_compress	= bitstream_read(bitstream, 4);
	    frame_info->gr[gr].ch[ch].window_switching_flag = bitstream_read_1(bitstream);
	    if (frame_info->gr[gr].ch[ch].window_switching_flag) {
		frame_info->gr[gr].ch[ch].block_type       = bitstream_read(bitstream, 2);
		frame_info->gr[gr].ch[ch].mixed_block_flag = bitstream_read_1(bitstream);
		for (region = 0; region < 2; region++)
		    frame_info->gr[gr].ch[ch].table_select[region] = 
			bitstream_read(bitstream, 5);
		for (window = 0; window < 3; window++)
		    frame_info->gr[gr].ch[ch].subblock_gain[window] =
			bitstream_read(bitstream, 3);
		
		block_type = frame_info->gr[gr].ch[ch].block_type;
		if ((block_type == 1) || ((block_type == 2) && frame_info->gr[gr].ch[ch].mixed_block_flag) || (block_type == 3))
		    frame_info->gr[gr].ch[ch].region0_count = 7;	// default value
		else if ((block_type == 2) && (!frame_info->gr[gr].ch[ch].mixed_block_flag)) 
		    frame_info->gr[gr].ch[ch].region0_count = 8;	// default value
		frame_info->gr[gr].ch[ch].region1_count	= 0; //20 - frame_info->gr[gr].ch[ch].region0_count;	// default value 36 
	    }
	    else {			
		for (region = 0; region < 3; region++)
		    frame_info->gr[gr].ch[ch].table_select[region] = 
			bitstream_read(bitstream, 5);
		frame_info->gr[gr].ch[ch].region0_count = bitstream_read(bitstream, 4);
		frame_info->gr[gr].ch[ch].region1_count	= bitstream_read(bitstream, 3);
		frame_info->gr[gr].ch[ch].block_type	= 0; 	// default value
	    }		
	    frame_info->gr[gr].ch[ch].preflag	         = bitstream_read_1(bitstream);
	    frame_info->gr[gr].ch[ch].scalefac_scale     = bitstream_read_1(bitstream);
	    frame_info->gr[gr].ch[ch].count1table_select = bitstream_read_1(bitstream);
	}
}

void frame_main_data(CT_Bitstream *bitstream, CT_MPEGInfo *frame_info, CT_Decoder *decoder)
{
    char	ch, gr, sfb, window, scalefac, region, scfsi_band;
    int		i;
    int		is[MAX_SB*MAX_SS]; 		// 576*4   = 2304 byte
    fixed_point	xr[MAX_CH][MAX_SB*MAX_SS];	// 2*576*4 = 4608 byte
    int		start_bit;
    CT_Channel	*channel_info;
    short	*samples_out;
    unsigned long start_time;
// #ifdef DEBUG_FRAMES
    FILE* file;
    #if DATA_WIDTH == 32
	  	file = fopen("debug_data/debug_frames_32.txt", OUTPUT_FILE_TYPE);
	  #else
	  	file = fopen("debug_data/debug_frames_16.txt", OUTPUT_FILE_TYPE);
	  #endif
// #endif

		int j, k, l, block_type;
		fixed_point *temp;
		fixed_point temp_data;

		short *temp1;



/*			#ifdef ARGUS_DEBUG
			start_time = APB.TIMER.TCR;
			#endif*/

    samples_out = (short*)malloc(576*2*2*2);	// 576 samples, 16 bit, 2 ch, 2 gr
   
    for (gr = 0; gr < 2; gr++) {
    	gr_count = gr;
			for (ch = 0; ch < frame_info->num_channels; ch++) {
				ch_count = ch;
	  	  channel_info = &frame_info->gr[gr].ch[ch];	    
			
	    	scalefac = channel_info->scalefac_compress;	// get the scalefac for this channel and granule
		    start_bit = bitstream_pos(bitstream);
			
		    if ((channel_info->window_switching_flag == 1) && (channel_info->block_type == 2)) {
					if (channel_info->mixed_block_flag) {
		  		  for (sfb = 0; sfb < 8; sfb++) { /* use slen1 */
							frame_info->scalefac_l[ch][sfb]	=
			  		  bitstream_read_mod(bitstream, slen[0][scalefac]);
				    }
				    for (sfb = 3; sfb < 12; sfb++) 
							for (window = 0; window < 3; window++) 
			  			  if (sfb < 6) { /* use slen1 */
									frame_info->scalefac_s[ch][sfb][window]	=
				  			  bitstream_read_mod(bitstream, slen[0][scalefac]);	
						    }
						    else { /* use slen2 */
									frame_info->scalefac_s[ch][sfb][window]	=
				  			  bitstream_read_mod(bitstream, slen[1][scalefac]);	
						    }
					}
					else {
				    for (sfb = 0; sfb < 12; sfb++) 
							for (window = 0; window < 3; window++) 
						    if (sfb < 6) { /* use slen1 */
									frame_info->scalefac_s[ch][sfb][window]	=
				  	  		bitstream_read_mod(bitstream, slen[0][scalefac]);	
					  	  }
					    	else {/* use slen2 */
									frame_info->scalefac_s[ch][sfb][window]	=
							    bitstream_read_mod(bitstream, slen[1][scalefac]);
						    }
					}
		    }
		    else {
					if ((frame_info->scfsi[ch][0] == 0) || (gr == 0))
				    for (sfb = 0; sfb < 6; sfb++) { /* use slen1 */
							frame_info->scalefac_l[ch][sfb]	=
					    bitstream_read_mod(bitstream, slen[0][scalefac]);
		    		}
					if ((frame_info->scfsi[ch][1] == 0) || (gr == 0))
				    for (sfb = 6; sfb < 11; sfb++) { /* use slen1 */
							frame_info->scalefac_l[ch][sfb]	=
					    bitstream_read_mod(bitstream, slen[0][scalefac]);
			    	}
					if ((frame_info->scfsi[ch][2] == 0) || (gr == 0))
			    	for (sfb = 11; sfb < 16; sfb++) { /* use slen2 */
							frame_info->scalefac_l[ch][sfb]	=
					    bitstream_read_mod(bitstream, slen[1][scalefac]);
		    		}
					if ((frame_info->scfsi[ch][3] == 0) || (gr == 0))
				    for (sfb = 16; sfb < 21; sfb++) { /* use slen2 */
							frame_info->scalefac_l[ch][sfb]	=
					    bitstream_read_mod(bitstream, slen[1][scalefac]);
			  	}
		    }
		    frame_info->scalefac_l[ch][21] = 0;	// not mentioned in the ISO specification, but needs to be done.
	  	  frame_info->scalefac_l[ch][22] = 0;
	    	for (window = 0; window < 3; window++)
					frame_info->scalefac_s[ch][12][window] = 0;

				if (channel_info->mixed_block_flag == 1) mixed_block_type_count++;

		    huffman_codebits(bitstream, frame_info, gr, ch, is, start_bit);	// decode the huffman data

				#if NEWDATA == 1
					int data;
		
					#if GEN_MIXED_BLOCK == 1
						if (channel_info->block_type == 2) channel_info->mixed_block_flag = 1;
					#endif

					if (start_frame_num == -1) {
						if (ch == 0) fscanf(left_data, "frame: %d", &start_frame_num);
						else fscanf(right_data, "frame: %d", &start_frame_num);
					}
					if (start_frame_num <= frame_count && start_frame_num != -1) {
// 						printf("Getting new data for frame %d gr %d ch %d\n", frame_count, gr, ch);
						if (ch == 0) fscanf(left_data, "%d", &data);
						else fscanf(right_data, "%d", &data);

						channel_info->count1 = data;

						for (j = 0; j < 576; j++) {
							if (ch == 0) fscanf(left_data, "%d", &data);
							else fscanf(right_data, "%d", &data);
							is[j] = data;
						}
					}
				#endif
	
				#if NEWDATA == 1
					if (start_frame_num <= frame_count && start_frame_num != -1) {
				#endif

				match_criteria = 1;

				#ifdef GEN_MEM
					// for printing input data for requantizer
					#if GEN_TYPE == 4
						if (ch == TARGET_CHANNEL || STEREO == 1) {	
				
							test_case = TARGET_TEST_CASE;

							#if RANDOM_GEN == 1
//							f_temp_print("// This is randomly generated\n");
// 							f_mem_print("// This is randomly generated\n");
// 							printf("// This is randomly generated\n");
					
// 							for (j = 0; j < MAX_SB*MAX_SS; j++) {
// 								is[j] = rand() % 16001 - 8000;
// 								if (is[j] > 1023) is[j] = (is[j]/8) * 8;
// 							}
					
								match_criteria = 0;
					
								if (frame_count >= START_FRAME) {
									int ms_stereo = (frame_info->header.mode == MPEG_MODE_JOINT_STEREO) && (frame_info->header.mode_extension & MPEG_MS_STEREO);
									int i_stereo = (frame_info->header.mode == MPEG_MODE_JOINT_STEREO) && (frame_info->header.mode_extension & MPEG_INTENSITY_STEREO);
									printf("MS: %d\n", ms_stereo);
									printf("IS: %d\n", i_stereo);
									printf("block type == %d\n", channel_info->block_type);
									printf("block type == 2? %d\n", (channel_info->window_switching_flag && (channel_info->block_type == 2)));
									printf("test_case = %d\n", test_case);
									switch(test_case) {
									case 0:
										// Pure MS, normal block
										if (ms_stereo == 1 && i_stereo == 0) {
											if (channel_info->block_type != 2) {
												match_criteria = 1;
												f_mem_print("// This is test case %d\n", test_case);
												printf("// This is test case %d\n", test_case);
											}
										}
									break;
									case 1:
										// Pure MS, short block
										if (ms_stereo == 1 && i_stereo == 0) {
											if (channel_info->window_switching_flag && channel_info->block_type == 2) {
											match_criteria = 1;
											f_mem_print("// This is test case %d\n", test_case);
											printf("// This is test case %d\n", test_case);
										}
									}
									break;
									case 2:
										// Pure IS, normal block
										if (ms_stereo == 0 && i_stereo == 1) {
											if (channel_info->block_type != 2) {
												match_criteria = 1;
												f_mem_print("// This is test case %d\n", test_case);
												printf("// This is test case %d\n", test_case);
											}
										}
									break;
									case 3:
										// Pure IS, short block
										if (ms_stereo == 0 && i_stereo == 1) {
											if (channel_info->window_switching_flag && channel_info->block_type == 2) {
												match_criteria = 1;
												f_mem_print("// This is test case %d\n", test_case);
												printf("// This is test case %d\n", test_case);
											}
										}
									break;
									case 4:
										// MS+IS, normal block
										if (ms_stereo == 1 && i_stereo == 1) {
											if (channel_info->block_type != 2) {
												match_criteria = 1;
												f_mem_print("// This is test case %d\n", test_case);
												printf("// This is test case %d\n", test_case);
											}
										}
									break;
									case 5:
										// MS+IS, short block
										if (ms_stereo == 1 && i_stereo == 1) {
											if (channel_info->window_switching_flag && channel_info->block_type == 2) {
												match_criteria = 1;
												f_mem_print("// This is test case %d\n", test_case);
												printf("// This is test case %d\n", test_case);
											}
										}
									break;
									default:
										exit(1);
									break;
									}
								}
							#endif
			
							if (match_criteria == 1) {
								printf("// This is frame %d, gr %d, ch %d\n", frame_count, gr_count, ch_count);
								f_mem_print("// This is frame %d, gr %d, ch %d\n", frame_count, gr_count, ch_count);
								f_mem_print("// mode: %d\n", frame_info->header.mode);
								f_mem_print("// mode ext: %d\n", frame_info->header.mode_extension);				
								f_mem_print("// block_type: %d\n", channel_info->block_type);
								f_mem_print("// window_switching_flag: %d\n", channel_info->window_switching_flag);
								f_mem_print("// mixed_block_flag: %d\n", channel_info->mixed_block_flag);
								f_mem_print("// scalefac_scale: %d\n", channel_info->scalefac_scale);
								f_mem_print("// preflag: %d\n", channel_info->preflag);
								f_mem_print("// global_gain: %d\n", channel_info->global_gain);
								f_mem_print("\n");
				
								f_mem_print("// Sample data (0-575)\n");
								for (j = 0; j < 576; j++) {
									f_mem_print_syn(is[j]);
								}
								f_mem_print("// Dummy (576-591)\n");								

								for (j = 576; j < 592; j++) {
									f_mem_print_syn(0);
								}

								f_mem_print("// Side info\n");
								f_mem_print("// For granule 0 (592)\n");
						    channel_info = &frame_info->gr[0].ch[ch];		    
								temp_data = ((channel_info->window_switching_flag & 0x00000001) << 13) // [13] window_switching_flag
										| ((channel_info->mixed_block_flag & 0x00000001) << 12) // [12] mixed_block_flag
										| ((channel_info->preflag & 0x00000001) << 11) // [11] pre_flag
										| ((channel_info->scalefac_scale & 0x00000001) << 10) // [10] scalefac_scale
										| ((channel_info->block_type & 0x00000003) << 8) // [9:8] block type
										| (channel_info->global_gain & 0x000000FF); // [7:0] global gain
								f_mem_print_syn(temp_data);
								f_mem_print("// Subblock_gain for granule 0 (593)\n");
								temp_data = ((channel_info->subblock_gain[2] & 0x00000007) << 6) // subblock_gain for window 2
										| ((channel_info->subblock_gain[1] & 0x00000007) << 3) // subblock_gain for window 1
										| (channel_info->subblock_gain[0] & 0x00000007); // subblock_gain for window 0
								f_mem_print_syn(temp_data);

								f_mem_print("// count1 for granule 0 (594)\n");
								f_mem_print_syn(channel_info->count1);
				
								f_mem_print("// dummy (595)\n");
								f_mem_print_syn(0);

								f_mem_print("// For granule 1 (596)\n");
						    channel_info = &frame_info->gr[1].ch[ch];
								temp_data = ((channel_info->window_switching_flag & 0x00000001) << 13) // [13] window_switching_flag
										| ((channel_info->mixed_block_flag & 0x00000001) << 12) // [12] mixed_block_flag
										| ((channel_info->preflag & 0x00000001) << 11) // [11] pre_flag
										| ((channel_info->scalefac_scale & 0x00000001) << 10) // [10] scalefac_scale
										| ((channel_info->block_type & 0x00000003) << 8) // [9:8] block type
										| (channel_info->global_gain & 0x000000FF); // [7:0] global gain
								f_mem_print_syn(temp_data);
								f_mem_print("// Subblock_gain for granule 1 (597)\n");
								temp_data = ((channel_info->subblock_gain[2] & 0x00000007) << 6) // subblock_gain for window 2
										| ((channel_info->subblock_gain[1] & 0x00000007) << 3) // subblock_gain for window 1
										| (channel_info->subblock_gain[0] & 0x00000007); // subblock_gain for window 0
								f_mem_print_syn(temp_data);

								f_mem_print("// count1 for granule 1 (598)\n");
								f_mem_print_syn(channel_info->count1);
				
						    channel_info = &frame_info->gr[gr].ch[ch];

								f_mem_print("// Dummy (599-639)\n");								
								for (j = 599; j < 640; j++) {
									f_mem_print_syn(0);
								}
				
								f_mem_print("// scale factor\n");
								f_mem_print("// short block or mixed_block (640-687)\n");				
								if ((channel_info->block_type == 2) || channel_info->mixed_block_flag) {
									for (j = 0; j < 3; j++) {
										f_mem_print("// window %d\n", j);				
										for (k = 0; k < 12; k++) {
											temp_data = (frame_info->scalefac_s[ch][k][j] & 0x0000FFFF);
											f_mem_print_syn(temp_data);
										}
										for (k = 12; k < 16; k++) {
											f_mem_print_syn(0);
										}
									}
									f_mem_print("// Dummy (688-703)\n");								
									for (k = 688; k < 704; k++) {
										f_mem_print_syn(0);
									}
								} else {
									f_mem_print("// Dummy for non-short block (640-703)\n");								
									for (j = 640; j < 704; j++) {
										f_mem_print_syn(0);
									}
								}
				
								if ((channel_info->block_type != 2) || ((channel_info->block_type == 2) && channel_info->mixed_block_flag)) {
									f_mem_print("// long block or mixed_block (704-724)\n");				
									for (j = 0; j < 21; j++) {
										temp_data = (frame_info->scalefac_l[ch][j] & 0x0000FFFF);
										f_mem_print_syn(temp_data);
									}
								} else {
									f_mem_print("// Dummy (704-724)\n");								
									for (j = 704; j < 725; j++) {
										f_mem_print_syn(0);
									}
								}
								f_mem_print("// Dummy for non-normal block (725-1023)\n");								
								for (k = 725; k < 1024; k++) {
									f_mem_print_syn(0);
								}
							}
						}
					#endif // endif GEN_TYPE == 4
				#endif // endif GEN_MEM
				
				#ifdef GEN_ETHERNET
					// for printing input data for requantizer
					#if GEN_TYPE == 4
						if (ch == TARGET_CHANNEL || STEREO == 1) {	
				
							test_case = TARGET_TEST_CASE;

							#if RANDOM_GEN == 1
//							f_temp_print("// This is randomly generated\n");
// 							f_mem_print("// This is randomly generated\n");
//		 					printf("// This is randomly generated\n");
					
// 							for (j = 0; j < MAX_SB*MAX_SS; j++) {
// 								is[j] = rand() % 16001 - 8000;
// 								if (is[j] > 1023) is[j] = (is[j]/8) * 8;
// 							}
					
								match_criteria = 0;
					
								if (frame_count >= START_FRAME) {
									int ms_stereo = (frame_info->header.mode == MPEG_MODE_JOINT_STEREO) && (frame_info->header.mode_extension & MPEG_MS_STEREO);
									int i_stereo = (frame_info->header.mode == MPEG_MODE_JOINT_STEREO) && (frame_info->header.mode_extension & MPEG_INTENSITY_STEREO);
// 								printf("MS: %d\n", ms_stereo);
// 								printf("IS: %d\n", i_stereo);
// 								printf("block type == %d\n", channel_info->block_type);
// 								printf("block type == 2? %d\n", (channel_info->window_switching_flag && (channel_info->block_type == 2)));
// 								printf("test_case = %d\n", test_case);
									switch(test_case) {
									case 0:
										// Pure MS, normal block
										if (ms_stereo == 1 && i_stereo == 0) {
											if (channel_info->block_type != 2) {
												match_criteria = 1;
												if (ETHERNET_COMMENT == 1)
													f_eth_print("// This is test case %d\n", test_case);
												printf("// This is test case %d\n", test_case);
											}
										}
									break;
									case 1:
										// Pure MS, short block
										if (ms_stereo == 1 && i_stereo == 0) {
											if (channel_info->window_switching_flag && channel_info->block_type == 2) {
												match_criteria = 1;
												if (ETHERNET_COMMENT == 1)
													f_eth_print("// This is test case %d\n", test_case);
												printf("// This is test case %d\n", test_case);
											}
										}
									break;
									case 2:
										// Pure IS, normal block
										if (ms_stereo == 0 && i_stereo == 1) {
											if (channel_info->block_type != 2) {
												match_criteria = 1;
												if (ETHERNET_COMMENT == 1)
													f_eth_print("// This is test case %d\n", test_case);
												printf("// This is test case %d\n", test_case);
											}
										}
									break;
									case 3:
										// Pure IS, short block
										if (ms_stereo == 0 && i_stereo == 1) {
											if (channel_info->window_switching_flag && channel_info->block_type == 2) {
												match_criteria = 1;
												if (ETHERNET_COMMENT == 1)
													f_eth_print("// This is test case %d\n", test_case);
												printf("// This is test case %d\n", test_case);
											}
										}
									break;
									case 4:
										// MS+IS, normal block
										if (ms_stereo == 1 && i_stereo == 1) {
											if (channel_info->block_type != 2) {
												match_criteria = 1;
												if (ETHERNET_COMMENT == 1)
													f_eth_print("// This is test case %d\n", test_case);
												printf("// This is test case %d\n", test_case);
											}
										}
									break;
									case 5:
										// MS+IS, short block
										if (ms_stereo == 1 && i_stereo == 1) {
											if (channel_info->window_switching_flag && channel_info->block_type == 2) {
												match_criteria = 1;
												if (ETHERNET_COMMENT == 1)
													f_eth_print("// This is test case %d\n", test_case);
												printf("// This is test case %d\n", test_case);
											}
										}
									break;
									default:
										exit(1);
									break;
									}
								}
							#endif // RANDOM_GEN == 1
			
							if (match_criteria == 1) {
								if (ch == 0) eth_fp = eth_left_fp;
								else eth_fp = eth_right_fp;
								if (ETHERNET_COMMENT == 1) {
									f_eth_print("// Data for requantizer\n");					
									printf("// This is frame %d, gr %d, ch %d\n", frame_count, gr_count, ch_count);
									f_eth_print("// This is frame %d, gr %d, ch %d\n", frame_count, gr_count, ch_count);
									f_eth_print("// mode: %d\n", frame_info->header.mode);
									f_eth_print("// mode ext: %d\n", frame_info->header.mode_extension);				
									f_eth_print("// block_type: %d\n", channel_info->block_type);
									f_eth_print("// window_switching_flag: %d\n", channel_info->window_switching_flag);
									f_eth_print("// mixed_block_flag: %d\n", channel_info->mixed_block_flag);
									f_eth_print("// scalefac_scale: %d\n", channel_info->scalefac_scale);
									f_eth_print("// preflag: %d\n", channel_info->preflag);
									f_eth_print("// global_gain: %d\n", channel_info->global_gain);
								}
								
								if (ETHERNET_COMMENT == 1)
									f_eth_print("// Sample data (0-575)\n");
								for (j = 0; j < 576; j++) {
									f_eth_print_data(is[j]);
								}
				
								if (ETHERNET_COMMENT == 1)
									f_eth_print("// Frequency (576)\n");
								temp_data = frame_info->header.sampling_frequency;
								f_eth_print_data(temp_data);

								if (ETHERNET_COMMENT == 1)
									f_eth_print("// Mode (577)\n");
								temp_data = frame_info->header.mode;
								f_eth_print_data(temp_data);
				
								if (ETHERNET_COMMENT == 1)
									f_eth_print("// Mode Extension (578)\n");
								temp_data = frame_info->header.mode_extension;
								f_eth_print_data(temp_data);

								if (ETHERNET_COMMENT == 1)
									f_eth_print("// Dummy (579-591)\n");								

								for (j = 579; j < 592; j++) {
									f_eth_print_data(0);
								}

								if (ETHERNET_COMMENT == 1)
									f_eth_print("// Side info\n");
								if (ETHERNET_COMMENT == 1)
									f_eth_print("// For granule 0 (592)\n");
						    channel_info = &frame_info->gr[0].ch[ch];		    
								temp_data = ((channel_info->window_switching_flag & 0x00000001) << 13) // [13] window_switching_flag
										| ((channel_info->mixed_block_flag & 0x00000001) << 12) // [12] mixed_block_flag
										| ((channel_info->preflag & 0x00000001) << 11) // [11] pre_flag
										| ((channel_info->scalefac_scale & 0x00000001) << 10) // [10] scalefac_scale
										| ((channel_info->block_type & 0x00000003) << 8) // [9:8] block type
										| (channel_info->global_gain & 0x000000FF); // [7:0] global gain
								f_eth_print_data(temp_data);

								if (ETHERNET_COMMENT == 1)
									f_eth_print("// Subblock_gain for granule 0 (593)\n");
								temp_data = ((channel_info->subblock_gain[2] & 0x00000007) << 6) // subblock_gain for window 2
										| ((channel_info->subblock_gain[1] & 0x00000007) << 3) // subblock_gain for window 1
										| (channel_info->subblock_gain[0] & 0x00000007); // subblock_gain for window 0
								f_eth_print_data(temp_data);

								if (ETHERNET_COMMENT == 1)
									f_eth_print("// count1 for granule 0 (594)\n");
								f_eth_print_data(channel_info->count1);
				
								if (ETHERNET_COMMENT == 1)
									f_eth_print("// dummy (595)\n");
								f_eth_print_data(0);
				
								if (ETHERNET_COMMENT == 1)
									f_eth_print("// For granule 1 (596)\n");
						    channel_info = &frame_info->gr[1].ch[ch];
								temp_data = ((channel_info->window_switching_flag & 0x00000001) << 13) // [13] window_switching_flag
										| ((channel_info->mixed_block_flag & 0x00000001) << 12) // [12] mixed_block_flag
										| ((channel_info->preflag & 0x00000001) << 11) // [11] pre_flag
										| ((channel_info->scalefac_scale & 0x00000001) << 10) // [10] scalefac_scale
										| ((channel_info->block_type & 0x00000003) << 8) // [9:8] block type
										| (channel_info->global_gain & 0x000000FF); // [7:0] global gain
								f_eth_print_data(temp_data);
								if (ETHERNET_COMMENT == 1)
									f_eth_print("// Subblock_gain for granule 1 (597)\n");
								temp_data = ((channel_info->subblock_gain[2] & 0x00000007) << 6) // subblock_gain for window 2
										| ((channel_info->subblock_gain[1] & 0x00000007) << 3) // subblock_gain for window 1
										| (channel_info->subblock_gain[0] & 0x00000007); // subblock_gain for window 0
								f_eth_print_data(temp_data);

								if (ETHERNET_COMMENT == 1)
									f_eth_print("// count1 for granule 1 (598)\n");
								f_eth_print_data(channel_info->count1);				

		    				channel_info = &frame_info->gr[gr].ch[ch];

								if (ETHERNET_COMMENT == 1)
									f_eth_print("// Dummy (599-639)\n");								
								for (j = 599; j < 640; j++) {
									f_eth_print_data(0);
								}
				
								if (ETHERNET_COMMENT == 1)
									f_eth_print("// scale factor\n");
								if (ETHERNET_COMMENT == 1)
									f_eth_print("// short block or mixed_block (640-687)\n");				
								if ((channel_info->block_type == 2) || channel_info->mixed_block_flag) {
									for (j = 0; j < 3; j++) {
										if (ETHERNET_COMMENT == 1)
											f_eth_print("// window %d\n", j);				
										for (k = 0; k < 12; k++) {
											temp_data = (frame_info->scalefac_s[ch][k][j] & 0x0000FFFF);
											f_eth_print_data(temp_data);
										}
										for (k = 12; k < 16; k++) {
											f_eth_print_data(0);
										}
									}
									if (ETHERNET_COMMENT == 1)
										f_eth_print("// Dummy (688-703)\n");								
									for (k = 688; k < 704; k++) {
										f_eth_print_data(0);
									}
								} else {
									if (ETHERNET_COMMENT == 1)
										f_eth_print("// Dummy for non-short block (640-703)\n");								
									for (j = 640; j < 704; j++) {
										f_eth_print_data(0);
									}
								}
				
								if ((channel_info->block_type != 2) || ((channel_info->block_type == 2) && channel_info->mixed_block_flag)) {
									if (ETHERNET_COMMENT == 1)
										f_eth_print("// long block or mixed_block (704-724)\n");				
									for (j = 0; j < 21; j++) {
										temp_data = (frame_info->scalefac_l[ch][j] & 0x0000FFFF);
										f_eth_print_data(temp_data);
									}
								} else {
									if (ETHERNET_COMMENT == 1)
										f_eth_print("// Dummy (704-724)\n");								
									for (j = 704; j < 725; j++) {
										f_eth_print_data(0);
									}
								}
								if (ETHERNET_COMMENT == 1)
									f_eth_print("// Dummy for non-normal block (725-1023)\n");								
								for (k = 725; k < 1024; k++) {
									f_eth_print_data(0);
								}
				
// 							f_eth_print("haha\n");
							}
						}
					#endif // GEN_TYPE == 4
				#endif // GEN_ETHERNET

	      	frame_requantize(frame_info, channel_info, gr, ch, is, xr[ch]);	// requantize all frequency bands

			    frame_reorder(frame_info, channel_info, xr[ch]);	// reorder short blocks
			#if NEWDATA == 1
				}
			#endif
			} // end of for (ch = 0; ch < frame_info->num_channels; ch++);
			//		stereo_processing(frame_info, xr, lr);									// decode stereo data to each channel
			#if NEWDATA == 1
			if (start_frame_num <= frame_count && start_frame_num != -1) {
			#endif
				#if NEW_CODE == 1
					stereo_processing_new(frame_info, gr, xr);
				#endif

				for (ch = 0; ch < frame_info->num_channels; ch++) {
					ch_count = ch;

	  		  channel_info = &frame_info->gr[gr].ch[ch];

					#ifdef GEN_MEM
						// for printing input data for Alias
						#if GEN_TYPE == 0
							if (ch == TARGET_CHANNEL || STEREO == 1) {
								temp = &xr[ch];
								if (ch == 0) eth_fp = eth_left_fp;
								else eth_fp = eth_right_fp;
	
								#if RANDOM_GEN == 1				
									if (ETHERNET_COMMENT == 1)
										f_mem_print("// This is randomly generated\n");
									printf("// This is randomly generated\n");
					
									for (j = 0; j < MAX_SB*MAX_SS; j++) {
										temp[j] = rand();
									}
									channel_info->block_type = rand() % 4;
// 								channel_info->block_type = block_type_sequence[random_block];
//				 				printf("haha: %d\n", block_type_sequence[random_block]);
//								random_block += 1;
								#endif // RANDOM_GEN == 1
								f_mem_print("// Data for antialias\n");
								f_mem_print("// This is frame %d, gr %d, ch %d, block_type = %d\n", frame_count, gr_count, ch_count, channel_info->block_type);

								if (channel_info->block_type != 2 ) {
									for (j = 0; j < 576; j++) {
										if (j % 9 == 0) f_mem_print("// This is sample %d-%d\n", j, j+8);
										f_mem_print_syn(temp[j]);
										if (j % 9 == 8) {
											for (k = 0; k < 6; k++) {
												f_mem_print_syn(channel_info->block_type);
											}
											f_mem_print_syn(channel_info->window_switching_flag);										
											f_mem_print_syn(channel_info->mixed_block_flag);										
										}
									}
								} else {
							   if (channel_info->window_switching_flag && (channel_info->block_type == 2) && !channel_info->mixed_block_flag ) {
									for (j = 0; j < 32; j++) {
										for (k = 0; k < 3; k++) {
											f_mem_print("// This is sample %d-%d\n", j*18+k*6, j*18+k*6+5);
											for (l = 0; l < 6; l++) {
												f_mem_print_syn(temp[l*3+k+18*j]);
											}
											for (l = 0; l < 2; l++) {
												f_mem_print_syn(channel_info->block_type);
											}
										}
										for (l = 0; l < 6; l++) {
											f_mem_print_syn(channel_info->block_type);
										}
										f_mem_print_syn(channel_info->window_switching_flag);										
										f_mem_print_syn(channel_info->mixed_block_flag);										
									}
									} else {
				  				  if (channel_info->window_switching_flag && (channel_info->block_type == 2) && channel_info->mixed_block_flag ) {
								    	for (j = 0; j < 36; j++) {
												if (j % 9 == 0) f_mem_print("// This is sample %d-%d\n", j, j+8);
												f_mem_print_syn(temp[j]);
												if (j % 9 == 8) {
													for (k = 0; k < 6; k++) {
														f_mem_print_syn(channel_info->block_type);
													}
													f_mem_print_syn(channel_info->window_switching_flag);										
													f_mem_print_syn(channel_info->mixed_block_flag);										
												}
											}
						
											for (j = 2; j < 32; j++) {
												for (k = 0; k < 3; k++) {
													f_mem_print("// This is sample %d-%d\n", j*18+k*6, j*18+k*6+5);
													for (l = 0; l < 6; l++) {
														f_mem_print_syn(temp[l*3+k+18*j]);
													}
													for (l = 0; l < 2; l++) {
														f_mem_print_syn(channel_info->block_type);
													}
												}
												for (l = 0; l < 6; l++) {
													f_mem_print_syn(channel_info->block_type);
												}
												f_mem_print_syn(channel_info->window_switching_flag);										
												f_mem_print_syn(channel_info->mixed_block_flag);										
											}
			  					  }
									}
								}
							}
						#endif
					#endif // GEN_MEM
				
					#ifdef GEN_ETHERNET
						// for printing input data for Alias
						#if GEN_TYPE == 0
							if (ch == TARGET_CHANNEL || STEREO == 1) {
								temp = &xr[ch];
								if (ch == 0) eth_fp = eth_left_fp;
								else eth_fp = eth_right_fp;
	
								#if RANDOM_GEN == 1				
									if (ETHERNET_COMMENT == 1)
										f_eth_print("// This is randomly generated\n");
									printf("// This is randomly generated\n");
					
									for (j = 0; j < MAX_SB*MAX_SS; j++) {
										temp[j] = rand();
									}
									channel_info->block_type = rand() % 4;
								#endif
								if (ETHERNET_COMMENT == 1) {
									f_eth_print("// Data for antialias\n");
									f_eth_print("// This is frame %d, gr %d, ch %d, block_type = %d, switching = %d, mixed_block=%d\n", 
										frame_count, gr_count, ch_count, channel_info->block_type, channel_info->window_switching_flag, channel_info->mixed_block_flag);
								}

								if (channel_info->block_type != 2 ) {
									for (j = 0; j < 576; j++) {
										if (ETHERNET_COMMENT == 1)
											if (j % 9 == 0) f_eth_print("// This is sample %d-%d\n", j, j+8);
										f_eth_print_data(temp[j]);
										if (j % 9 == 8) {
											for (k = 0; k < 6; k++) {
												f_eth_print_data(channel_info->block_type);
											}
											f_eth_print_data(channel_info->window_switching_flag);										
											f_eth_print_data(channel_info->mixed_block_flag);										
										}
									}	
								} else {
							    if (channel_info->window_switching_flag && (channel_info->block_type == 2) && !channel_info->mixed_block_flag ) {
										for (j = 0; j < 32; j++) {
											for (k = 0; k < 3; k++) {
												if (ETHERNET_COMMENT == 1)
													f_eth_print("// This is sample %d-%d\n", j*18+k*6, j*18+k*6+5);
	
												for (l = 0; l < 6; l++) {
													f_eth_print_data(temp[l*3+k+18*j]);
												}
												for (l = 0; l < 2; l++) {
													f_eth_print_data(channel_info->block_type);
												}
											}
											for (l = 0; l < 6; l++) {
												f_eth_print_data(channel_info->block_type);
											}
											f_eth_print_data(channel_info->window_switching_flag);										
											f_eth_print_data(channel_info->mixed_block_flag);										
										}
									} else {
			    					if (channel_info->window_switching_flag && (channel_info->block_type == 2) && channel_info->mixed_block_flag ) {
								    	for (j = 0; j < 36; j++) {
												if (ETHERNET_COMMENT == 1)
													if (j % 9 == 0) f_eth_print("// This is sample %d-%d\n", j, j+8);
												f_eth_print_data(temp[j]);
												if (j % 9 == 8) {
													for (k = 0; k < 6; k++) {
														f_eth_print_data(channel_info->block_type);
													}
													f_eth_print_data(channel_info->window_switching_flag);										
													f_eth_print_data(channel_info->mixed_block_flag);										
												}
											}
							
											for (j = 2; j < 32; j++) {
												for (k = 0; k < 3; k++) {
													if (ETHERNET_COMMENT == 1)
														f_eth_print("// This is sample %d-%d\n", j*18+k*6, j*18+k*6+5);

													for (l = 0; l < 6; l++) {
														f_eth_print_data(temp[l*3+k+18*j]);
													}
													for (l = 0; l < 2; l++) {
														f_eth_print_data(channel_info->block_type);
													}
												}
												for (l = 0; l < 6; l++) {
													f_eth_print_data(channel_info->block_type);
												}
												f_eth_print_data(channel_info->window_switching_flag);										
												f_eth_print_data(channel_info->mixed_block_flag);										
											}
			    					}
									}
								}
							}
						#endif
					#endif // GEN_ETHERNET

			    alias_reduction(xr[ch], is, channel_info);
					
					#ifdef GEN_MEM
						// for printing input data for IMDCT
						#if GEN_TYPE == 1
							if (ch == TARGET_CHANNEL || STEREO == 1) {
								f_mem_print("// This is frame %d, gr %d, ch %d\n", frame_count, gr, ch);
// 							for (k = 0; k < 32; k++) {
									k = 0;
									temp = &is[k*MAX_SS];
									for (j = 0; j < MAX_SS; j++) {
										f_mem_print_syn(temp[j]);
									}
//			 				}
							}
						#endif
					#endif // GEN_MEM    

			    imdct(is, xr[ch], ch, channel_info);	// IMDCT

//				frequency_inversion(xr[ch]);										// frequency inversion

					#ifdef IS_HARDWARE_ACCELERATED
				    subband_synthesis_hw(xr[ch], ch, gr, &samples_out[gr*576*2], frame_info, decoder);	// subbandsynthesis (output PCM samples)
					#else

						#ifdef GEN_MEM
							// for printing input data for subsyn
							#if GEN_TYPE == 2
								if (ch == TARGET_CHANNEL || STEREO == 1) {
									f_mem_print("// This is frame %d, gr %d, ch %d\n", frame_count, gr, ch);
									for (j = 0; j < MAX_SS; j++) {
										temp = &xr[ch][j*MAX_SB];
										for (k = 0; k < 32; k++) {
											f_mem_print_syn(temp[k]);
										}
									}
								}
							#endif
						#endif // GEN_MEM

				    subband_synthesis(xr[ch], ch, &samples_out[gr*576*2]);				// subbandsynthesis (output PCM samples)

						#ifdef GEN_MEM
							// for printing input data for AC97
							#if GEN_TYPE == 3
								if (ch == TARGET_CHANNEL || STEREO == 1) {
									f_mem_print("// This is frame %d, gr %d, ch %d\n", frame_count, gr, ch);
									temp = &samples_out[gr*576*2];
									for (j = 0; j < MAX_SB*MAX_SS; j++)
										f_mem_print_syn(temp[2*j]);
								}
							#endif
							if (mem_count > NUM_MEM) exit(1);
						#endif // GEN_MEM
					#endif // IS_HARDWARE_ACCELERATED		

					temp1 = &samples_out[gr*576*2];
					for (j = 0; j < MAX_SB*MAX_SS; j++) {
						decoder->pcm_sample[gr][ch][j/MAX_SB][j%MAX_SB] = temp1[2*j+ch];
// 					printf("i=%d, j=%d, #=%d, l=%d, %04X\n", j%MAX_SS, j/MAX_SS, j, ch, decoder->pcm_sample[gr][ch][j%MAX_SS][j/MAX_SS] & 0x0000FFFF);					
					}							
				} // for (ch = 0; ch < frame_info->num_channels; ch++);
	  	#if NEWDATA == 1
			}
			#endif
 		} // for (gr = 0; gr < 2; gr++)

/*			#ifdef ARGUS_DEBUG
		start_time = APB.TIMER.TCR - start_time;
		printf("DECODER : took %d us\n", start_time);
		#endif*/

		#ifndef IS_HARDWARE_ACCELERATED
	    if (frame_info->num_channels == 1)
				for(i = 0; i < 576*2; i++)
			    samples_out[(i << 1) + 1] = samples_out[i << 1];

	    buffer_add(decoder->output_buffer, (unsigned char*)samples_out, 576*2*2*2, 0, 0);
		#endif
    // ancillary data

		#ifdef DEBUG_FRAMES
			// 		printf("printing debug frome info...\n");
   		fprintf(file, "Start of frame ------------------------------------------------\n");
		  fprintf(file, "ID:            %d\n", frame_info->header.ID);
  	  fprintf(file, "Layer:         %d\n", frame_info->header.layer);
   		fprintf(file, "Protection:    %d\n", frame_info->header.protection);
		  fprintf(file, "Bitrate index: %d\n", frame_info->header.bitrate_index);
  	  fprintf(file, "Frequency:     %d\n", frame_info->header.sampling_frequency);
  		fprintf(file, "Padding:       %d\n", frame_info->header.padding_bit);
		  fprintf(file, "Private:       %d\n", frame_info->header.private_bit);
  	  fprintf(file, "Mode:          %d\n", frame_info->header.mode);
  		fprintf(file, "Mode ext:      %d\n", frame_info->header.mode_extension);
		  fprintf(file, "Copyright:     %d\n", frame_info->header.copyright);
  	  fprintf(file, "Original:      %d\n", frame_info->header.original);
  		fprintf(file, "Emphasis:      %d\n", frame_info->header.emphasis);
		  fprintf(file, "CRC:           %d\n", frame_info->crc);
  	  fprintf(file, "\n");
	
  		fprintf(file, "Main data begin:  %d\n", frame_info->main_data_begin);
	  	fprintf(file, "Private bits:     %d\n", frame_info->private_bits);
	    for (ch = 0; ch < frame_info->num_channels; ch++) 
				for (scfsi_band = 0; scfsi_band < 4; scfsi_band++) 
	  			fprintf(file, "scsfi[ch=%d][band=%d]:               %d\n", ch, scfsi_band, frame_info->scfsi[ch][scfsi_band]);
		    for (gr = 0; gr < 2; gr++) 
					for (ch = 0; ch < frame_info->num_channels; ch++) {
				    fprintf(file, "\npart2_3_length[ch=%d][gr=%d]:        %d\n", ch, gr, frame_info->gr[gr].ch[ch].part2_3_length);
				    fprintf(file, "big_values[ch=%d][gr=%d]:            %d\n", ch, gr, frame_info->gr[gr].ch[ch].big_values);
			  	  fprintf(file, "global_gain[ch=%d][gr=%d]:           %d\n", ch, gr, frame_info->gr[gr].ch[ch].global_gain);
				    fprintf(file, "scalefac_compress[ch=%d][gr=%d]:     %d\n", ch, gr, frame_info->gr[gr].ch[ch].scalefac_compress);
				    fprintf(file, "window_switching_flag[ch=%d][gr=%d]: %d\n", ch, gr, frame_info->gr[gr].ch[ch].window_switching_flag);
				    fprintf(file, "block_type[ch=%d][gr=%d]:            %d\n", ch, gr, frame_info->gr[gr].ch[ch].block_type);
				    fprintf(file, "mixed_block_flag[ch=%d][gr=%d]:      %d\n", ch, gr, frame_info->gr[gr].ch[ch].mixed_block_flag);
			  	  for (window = 0; window < 3; window++)
							fprintf(file, "subblock_gain[ch=%d][gr=%d][wnd=%d]:  %d\n", ch, gr, window, frame_info->gr[gr].ch[ch].subblock_gain[window]);
				    for (region = 0; region < 3; region++)
							fprintf(file, "table_select[ch=%d][gr=%d][reg=%d]:   %d\n", ch, gr, region, frame_info->gr[gr].ch[ch].table_select[region]);
				    fprintf(file, "region0_count[ch=%d][gr=%d]:         %d\n", ch, gr, frame_info->gr[gr].ch[ch].region0_count);
	  				fprintf(file, "region1_count[ch=%d][gr=%d]:         %d\n", ch, gr, frame_info->gr[gr].ch[ch].region1_count);
			  	  fprintf(file, "preflag[ch=%d][gr=%d]:               %d\n", ch, gr, frame_info->gr[gr].ch[ch].preflag);
	  				fprintf(file, "scalefac_scale[ch=%d][gr=%d]:        %d\n", ch, gr, frame_info->gr[gr].ch[ch].scalefac_scale);
				    fprintf(file, "count1table_select[ch=%d][gr=%d]:    %d\n", ch, gr, frame_info->gr[gr].ch[ch].count1table_select);
				}
  		fprintf(file, "\n");
			fclose(file);
		#endif
}
