// -------------------------------------------------------
//
// Author: Thomas Lenart    
// Email:  thomas.lenart@es.lth.se
// Date:   March 2003
//
// -------------------------------------------------------
#include "main.h"

#ifdef DEBUG_HUFFMAN
#include <stdio.h>
#endif

#include "huffman.h"

static struct _huffman_table {
	unsigned char const *table[32];
} huffman_table;

static const unsigned char linbit_table[] =	{ 
												#include "linbits" 
											};
static const unsigned char huffmanA[] =		{ 
												#include "HuffmanA" 
											};
static const unsigned char huffman0[] =		{ 
												#include "Huffman0" 
											};
static const unsigned char huffman1[] =		{ 
												#include "Huffman1" 
											};
static const unsigned char huffman2[] =		{ 
												#include "Huffman2" 
											};
static const unsigned char huffman3[] =		{ 
												#include "Huffman3" 
											};
static const unsigned char huffman5[] =		{ 
												#include "Huffman5" 
											};
static const unsigned char huffman6[] =		{ 
												#include "Huffman6" 
											};
static const unsigned char huffman7[] =		{ 
												#include "Huffman7" 
											};
static const unsigned char huffman8[] =		{ 
												#include "Huffman8" 
											};
static const unsigned char huffman9[] =		{ 
												#include "Huffman9" 
											};
static const unsigned char huffman10[] =	{ 
												#include "Huffman10" 
											};
static const unsigned char huffman11[] =	{ 
												#include "Huffman11" 
											};
static const unsigned char huffman12[] =	{ 
												#include "Huffman12" 
											};
static const unsigned char huffman13[] =	{ 
												#include "Huffman13" 
											};
static const unsigned char huffman15[] =	{ 
												#include "Huffman15" 
											};
static const unsigned char huffman16[] =	{ 
												#include "Huffman16" 
											};
static const unsigned char huffman24[] =	{ 
												#include "Huffman24" 
											};

/*

		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx111111111111111111111111000000000000000000000000
		Bigvalues * 2								Count1					rzero
		00000000000000001111111111111122222222222222
		region0_count	region1_count

*/

void huffman_init(void)
{
	huffman_table.table[0] = huffman0;
	huffman_table.table[1] = huffman1;
	huffman_table.table[2] = huffman2;
	huffman_table.table[3] = huffman3;
	huffman_table.table[5] = huffman5;
	huffman_table.table[6] = huffman6;
	huffman_table.table[7] = huffman7;
	huffman_table.table[8] = huffman8;
	huffman_table.table[9] = huffman9;
	huffman_table.table[10] = huffman10;
	huffman_table.table[11] = huffman11;
	huffman_table.table[12] = huffman12;
	huffman_table.table[13] = huffman13;
	huffman_table.table[15] = huffman15;
	huffman_table.table[16] = huffman16;
	huffman_table.table[17] = huffman16;
	huffman_table.table[18] = huffman16;
	huffman_table.table[19] = huffman16;
	huffman_table.table[20] = huffman16;
	huffman_table.table[21] = huffman16;
	huffman_table.table[22] = huffman16;
	huffman_table.table[23] = huffman16;
	huffman_table.table[24] = huffman24;
	huffman_table.table[25] = huffman24;
	huffman_table.table[26] = huffman24;
	huffman_table.table[27] = huffman24;
	huffman_table.table[28] = huffman24;
	huffman_table.table[29] = huffman24;
	huffman_table.table[30] = huffman24;
	huffman_table.table[31] = huffman24;
}

void huffman_codebits(CT_Bitstream *bitstream, CT_MPEGInfo *frame_info, char gr, char ch, int is[MAX_SB*MAX_SS], int start_bit)
{
	unsigned int			line, region, old_region;
	int						v, w, x, y, linbits, count1;
	unsigned int			value, index, region1start, region2start;
	unsigned int tmp_value;
	const unsigned char*	table;
	CT_Channel				*channel_info = &frame_info->gr[gr].ch[ch];
  //ddxo
  int main_data_start = start_bit;
  int main_data_end = start_bit + channel_info->part2_3_length;
	#ifdef DEBUG_HUFFMAN
	FILE* file;
	#if DATA_WIDTH == 32
		file = fopen("debug_data/debug_huffman_32.txt", OUTPUT_FILE_TYPE);
	#else
		file = fopen("debug_data/debug_huffman_16.txt", OUTPUT_FILE_TYPE);
	#endif
	#endif
	
	old_region = 3;		// start with a non valid region

	if ((channel_info->window_switching_flag) && (channel_info->block_type == 2)) {    
		region1start = 36;  
		region2start = 576; 
	}
	else {          
		region1start = sfBandIndex_l[frame_info->header.sampling_frequency][channel_info->region0_count + 1];
  		if (frame_info->gr[gr].ch[ch].window_switching_flag)
			region2start = 576; 
  		else
			region2start = sfBandIndex_l[frame_info->header.sampling_frequency][channel_info->region0_count + channel_info->region1_count + 2];
	}

	//--------------------------
	// Decode big_values
	//--------------------------
	
	for (line = 0; line < (channel_info->big_values << 1); line += 2) {	
		
		if (line < region1start) 
			region = 0;
		else if (line < region2start) 
			region = 1;
		else
			region = 2;

		if (region != old_region) {
			table = huffman_table.table[channel_info->table_select[region]];		// get the huffman table
			linbits = linbit_table[channel_info->table_select[region]];
			old_region = region;
		}

		index = 0;
		do {
			if (table[index] == 0) {						// if end of table
				value = table[index + 1];					// get value
				break;		
			}
			if (bitstream_read_1_mod(bitstream))
				index += (table[index] + 1) << 1;			// if 1, find 1 node
			else
				index += table[index] << 1;					// if 0, find 0 node
		} while (1);
	
		x = value >> 4;
		y = value & 0x0F;

		if ((x == 15) && (linbits > 0))						// check if linbits should be used
			x += bitstream_read_mod(bitstream, linbits);	// if so, read linbits from the stream
		if (x)												// if x > 0, then read sign bit
			if (bitstream_read_1_mod(bitstream))			
				x = -x;
		if ((y == 15) && (linbits > 0))
			y += bitstream_read_mod(bitstream, linbits);
		if (y) 
			if (bitstream_read_1_mod(bitstream))
				y = -y;

		if (x > 8191)
			x = 8191;
		if (y > 8191)
			y = 8191;

		#ifdef DEBUG_HUFFMAN
/*		fprintf(file, "huffman[x] %d\n", x);	
		fprintf(file, "huffman[y] %d\n", y);*/
		fprintf(file, "%d\n", x);	
		fprintf(file, "%d\n", y);	
		#endif

		is[line] = x;
		is[line+1] = y;
	}
	count1 = channel_info->part2_3_length - (bitstream_pos(bitstream) - start_bit);		// number of bits left in main data
	table = huffmanA;																	// get the huffman table for A
	channel_info->count1 = 0;

	//--------------------------
	// Decode rcount1
	//--------------------------

	for (; (line < 576) && (count1 > 0); line += 4) {

		start_bit = bitstream_pos(bitstream);
		
		if (channel_info->count1table_select == 0) {		// quadruples table A
			index = 0;
			do {
				if (table[index] == 0) {					// if end of table
					value = table[index + 1];				// get value
					break;		
				}
				if (bitstream_read_1_mod(bitstream))
					index += (table[index] + 1) << 1;		// if 1, find 1 node
				else
					index += table[index] << 1;				// if 0, find 0 node
			} while (1);
		}
		else {												// quadruples table B
			value = 15 - bitstream_read_mod(bitstream, 4);	// invert bits from stream
		}
		
// 		if (value) {
// 			channel_info->count1+=1;
// 			tmp_value = value;
// 		}
	
		v = value >> 3;										// extract the v, w, x, y values 
		w = (value >> 2) & 1;
		x = (value >> 1) & 1;
		y = value & 1;

		if (v)
			if (bitstream_read_1_mod(bitstream))
				v = -v;
		if (w) 
			if (bitstream_read_1_mod(bitstream))
				w = -w;
		if (x) 
			if (bitstream_read_1_mod(bitstream))
				x = -x;
		if (y)
			if (bitstream_read_1_mod(bitstream))
				y = -y;

		count1 -= bitstream_pos(bitstream) - start_bit;

		#ifdef DEBUG_HUFFMAN
/*		fprintf(file, "huffman[v] %d\n", v);	
		fprintf(file, "huffman[w] %d\n", w);	
		fprintf(file, "huffman[x] %d\n", x);	
		fprintf(file, "huffman[y] %d\n", y);*/
		fprintf(file, "%d\n", v);	
		fprintf(file, "%d\n", w);	
		fprintf(file, "%d\n", x);	
		fprintf(file, "%d\n", y);	
		#endif

		is[line+0] = v;
		is[line+1] = w;
		is[line+2] = x;
		is[line+3] = y;
	}
//ddxo
  bitstream_read_mod(bitstream, main_data_end - bitstream_pos(bitstream));

// 	channel_info->count1 = (channel_info->count1+1) << 2;
	channel_info->count1 = line;
		
// 	if (tmp_value & 3 == 0) {
// 		channel_info->count1 -= 2;
// 	}
	
	for (; line < 576; line++) {
		is[line] = 0;			// zeros. pad with 0
		#ifdef DEBUG_HUFFMAN
		fprintf(file, "0\n");	
		#endif
    }

	#ifdef DEBUG_HUFFMAN
	fclose(file);
	#endif
}
