// -------------------------------------------------------
//
// Author: Thomas Lenart    
// Email:  thomas.lenart@es.lth.se
// Date:   March 2003
//
// -------------------------------------------------------
#include "main.h"

#include <stdio.h>

#include "memhandler.h"
#include "mpeg.h"
#include "frames.h"
#include "huffman.h"
#include "imdct.h"
#include "subbandsyn.h"

static CT_MPEGInfo	frame_info;

CT_Decoder *mpeg_stream_init(int cbuffer_size)			// cbuffer_size = 2^N
{
	CT_Decoder *mpeg = (CT_Decoder*)malloc(sizeof(CT_Decoder));

	mpeg->input_buffer = buffer_create();		// create an input buffer
	mpeg->output_buffer = buffer_create();		// create an output buffer
	mpeg->bitstream = bitstream_create(0, 0);	// create a new bitstream from input_buffer
	if (cbuffer_size > 0) {
		mpeg->cbuffer = (unsigned char*)malloc(cbuffer_size);		     // alloc a fixed block
		mpeg->cb_bitstream = bitstream_create(mpeg->cbuffer, cbuffer_size);  // create a new bitstream from input_buffer
	}
	mpeg->cb_pos = 0;
	mpeg->cb_size = cbuffer_size;
	
	huffman_init();	// initiate the huffman tables
	imdct_init();
	subband_synthesis_init();
	stereo_init();
	
	// initiate all the files
	#ifdef DEBUG
		FILE *file;
    #if DATA_WIDTH == 16
      file = fopen("decoded_samples.txt", INIT_OUTPUT_FILE_TYPE);
      fclose(file);
      file = fopen("requantized_samples.txt", INIT_OUTPUT_FILE_TYPE);
      fclose(file);
      file = fopen("stereo_samples.txt", INIT_OUTPUT_FILE_TYPE);
      fclose(file);
    #endif
		#if DATA_WIDTH == 32
  	 file = fopen("debug_data/debug_frames_32.txt", INIT_OUTPUT_FILE_TYPE);
	  #else
  	 file = fopen("debug_data/debug_frames_16.txt", INIT_OUTPUT_FILE_TYPE);
	  #endif
	  fprintf(file, "// New_code %d, New_Mult %d\n", NEW_CODE, NEW_MULT);
	  fclose(file);
		#if DATA_WIDTH == 32
			file = fopen("debug_data/debug_huffman_32.txt", INIT_OUTPUT_FILE_TYPE);
		#else
			file = fopen("debug_data/debug_huffman_16.txt", INIT_OUTPUT_FILE_TYPE);
		#endif
		fclose(file);
		#if DATA_WIDTH == 32
			file = fopen("debug_data/debug_requantize_input_32.txt", INIT_OUTPUT_FILE_TYPE);
		#else
			file = fopen("debug_data/debug_requantize_input_16.txt", INIT_OUTPUT_FILE_TYPE);
		#endif
		fclose(file);
		#if DATA_WIDTH == 32
			file = fopen("debug_data/debug_requantize_output_32.txt", INIT_OUTPUT_FILE_TYPE);
		#else
			file = fopen("debug_data/debug_requantize_output_16.txt", INIT_OUTPUT_FILE_TYPE);
		#endif
		fclose(file);
		#if DATA_WIDTH == 32
			file = fopen("debug_data/debug_requantize_intermediate_32.txt", INIT_OUTPUT_FILE_TYPE);
		#else
			file = fopen("debug_data/debug_requantize_intermediate_16.txt", INIT_OUTPUT_FILE_TYPE);
		#endif
		fclose(file);

		#if DATA_WIDTH == 32
			file = fopen("debug_data/debug_reorder_32.txt", INIT_OUTPUT_FILE_TYPE);
		#else
			file = fopen("debug_data/debug_reorder_16.txt", INIT_OUTPUT_FILE_TYPE);
		#endif
		fclose(file);
		#if DATA_WIDTH == 32
			file = fopen("debug_data/debug_stereo_input_32.txt", INIT_OUTPUT_FILE_TYPE);
		#else
			file = fopen("debug_data/debug_stereo_input_16.txt", INIT_OUTPUT_FILE_TYPE);
		#endif
		fclose(file);
		#if DATA_WIDTH == 32
			file = fopen("debug_data/debug_stereo_intermediate_32.txt", INIT_OUTPUT_FILE_TYPE);
		#else
			file = fopen("debug_data/debug_stereo_intermediate_16.txt", INIT_OUTPUT_FILE_TYPE);
		#endif
		fclose(file);
		#if DATA_WIDTH == 32
			file = fopen("debug_data/debug_stereo_output_32.txt", INIT_OUTPUT_FILE_TYPE);
		#else
			file = fopen("debug_data/debug_stereo_output_16.txt", INIT_OUTPUT_FILE_TYPE);
		#endif
		fclose(file);
	  #if DATA_WIDTH == 32
			file = fopen("debug_data/debug_alias_input_32.txt", INIT_OUTPUT_FILE_TYPE);
		#else
			file = fopen("debug_data/debug_alias_input_16.txt", INIT_OUTPUT_FILE_TYPE);
		#endif
    fclose(file);
    #if DATA_WIDTH == 32
			file = fopen("debug_data/debug_alias_output_32.txt", INIT_OUTPUT_FILE_TYPE);
		#else
			file = fopen("debug_data/debug_alias_output_16.txt", INIT_OUTPUT_FILE_TYPE);
		#endif
		fclose(file);
		
		#if DATA_WIDTH == 32
    file = fopen("debug_data/debug_imdct_32.txt", INIT_OUTPUT_FILE_TYPE);
    #else
    file = fopen("debug_data/debug_imdct_16.txt", INIT_OUTPUT_FILE_TYPE);
    #endif
    fclose(file);
    
 		#if DATA_WIDTH == 32
    file = fopen("debug_data/debug_icg_b4_window_32.txt", INIT_OUTPUT_FILE_TYPE);
    #else
    file = fopen("debug_data/debug_icg_b4_window_16.txt", INIT_OUTPUT_FILE_TYPE);
    #endif
    fclose(file);

 		#if DATA_WIDTH == 32
    file = fopen("debug_data/debug_ipg_b4_window_32.txt", INIT_OUTPUT_FILE_TYPE);
    #else
    file = fopen("debug_data/debug_ipg_b4_window_16.txt", INIT_OUTPUT_FILE_TYPE);
    #endif
    fclose(file);

 		#if DATA_WIDTH == 32
    file = fopen("debug_data/debug_icg_32.txt", INIT_OUTPUT_FILE_TYPE);
    #else
    file = fopen("debug_data/debug_icg_16.txt", INIT_OUTPUT_FILE_TYPE);
    #endif
    fclose(file);

 		#if DATA_WIDTH == 32
    file = fopen("debug_data/debug_ipg_32.txt", INIT_OUTPUT_FILE_TYPE);
    #else
    file = fopen("debug_data/debug_ipg_16.txt", INIT_OUTPUT_FILE_TYPE);
    #endif
    fclose(file);
		
		#if DATA_WIDTH == 32
		file = fopen("debug_data/debug_to_subsyn_32.txt", INIT_OUTPUT_FILE_TYPE);
		#else
		file = fopen("debug_data/debug_to_subsyn_16.txt", INIT_OUTPUT_FILE_TYPE);
		#endif
		fclose(file);
		#if DATA_WIDTH == 32
		file = fopen("debug_data/debug_from_subsyn_32.txt", INIT_OUTPUT_FILE_TYPE);
		#else
		file = fopen("debug_data/debug_from_subsyn_16.txt", INIT_OUTPUT_FILE_TYPE);
		#endif
		fclose(file);
		#if DATA_WIDTH == 32
		file = fopen("c_32.txt", INIT_OUTPUT_FILE_TYPE);
		#else
		file = fopen("d_16.txt", INIT_OUTPUT_FILE_TYPE);
		#endif
		fclose(file);
		#if DATA_WIDTH == 32
		file = fopen("debug_data/debug_mdct_input_32.txt", INIT_OUTPUT_FILE_TYPE);
		#else
		file = fopen("debug_data/debug_mdct_input_16.txt", INIT_OUTPUT_FILE_TYPE);
		#endif
		fclose(file);
		#if DATA_WIDTH == 32
		file = fopen("debug_data/debug_mdct_intermediate_32.txt", INIT_OUTPUT_FILE_TYPE);
		#else
		file = fopen("debug_data/debug_mdct_intermediate_16.txt", INIT_OUTPUT_FILE_TYPE);
		#endif
		fclose(file);

		#if DATA_WIDTH == 32
		file = fopen("debug_data/debug_mdct_output_32.txt", INIT_OUTPUT_FILE_TYPE);
		#else
		file = fopen("debug_data/debug_mdct_output_16.txt", INIT_OUTPUT_FILE_TYPE);
		#endif
		fclose(file);	
	#endif

	return mpeg;
}

CT_MPEGInfo* mpeg_stream_decode(CT_Decoder *decoder)
{
	unsigned long start_time;
	#ifdef DEBUG_MPEG
	FILE* file;
	int gr, ch;
	static counter = 0;
	#if DATA_WIDTH == 32
	file = fopen("debug_data/debug_mpeg_32.txt", OUTPUT_FILE_TYPE);
	#else
	file = fopen("debug_data/debug_mpeg_16.txt", OUTPUT_FILE_TYPE);
	#endif
	#endif

	#ifdef ARGUS_DEBUG
	{
		unsigned long size, used, avail;
		kheap_getsizes(&size, &used, &avail);	
		printf("DECODER : free heap size %d\n", avail);
		start_time = APB.TIMER.TCR;
	}
	#endif

	bitstream_init(decoder->bitstream, decoder->input_buffer->first->data, 
		       decoder->input_buffer->first->data_size);	

	if (find_frame_sync(decoder->bitstream) != -1) {
		if (frame_header(decoder->bitstream, &frame_info)) {
			if (frame_error_check(decoder->bitstream, &frame_info)) {
				frame_audio_data(decoder->bitstream, &frame_info);

				bitstream_set_pos(decoder->cb_bitstream, 
						  ((decoder->input_buffer->first->cb_offset -
						    frame_info.main_data_begin) % decoder->cb_size) << 3);

				frame_main_data(decoder->cb_bitstream, &frame_info, decoder);
				/* frame_ancillary_data(); */
			}
		}
	}

	#ifdef ARGUS_DEBUG
	start_time = APB.TIMER.TCR - start_time;
	printf("DECODER : decoding took %d us\n", start_time);
	#endif

	#ifdef DEBUG_MPEG
	fprintf(file, "%d ---------------\n", counter * 32 * 18 * 4);
	fprintf(file, "mode=%s, mode_ext=%s\n", mode_names[frame_info.header.mode],
		mode_ext_names[frame_info.header.mode_extension]);
	for (gr = 0; gr < 2; gr++)
		for (ch = 0; ch < 2; ch++)
			fprintf(file, "block_type[gr=%d][ch=%d]=%d\n", gr, ch, frame_info.gr[gr].ch[ch].block_type);
	fprintf(file, "\n");
	fclose(file);
	counter++;
	#endif

	buffer_delete(decoder->input_buffer, CT_DELETE_DATA);

	frame_info.frame_duration = 26122;	// us per frame

	return &frame_info;
}
