#ifndef __MPEG_H__
#define __MPEG_H__

#include "constants.h"
#include "decoder.h"

typedef struct {
    unsigned char	ID;
    unsigned char	layer;
    unsigned char	protection;
    unsigned char	bitrate_index;
    unsigned char	sampling_frequency;
    unsigned char	padding_bit;
    unsigned char	private_bit;
    unsigned char	mode;
    unsigned char	mode_extension;
    unsigned char	copyright;
    unsigned char	original;
    unsigned char	emphasis;
} CT_MPEGHeader;

typedef struct {
    CT_MPEGHeader	header;
    unsigned short	crc;			// 16 bit CRC
    unsigned short	main_data_begin;	// 9 bit negative offset, start of main data
    unsigned char	private_bits;		// 3 bits in stereo, 5 bits in mono
    unsigned char	scfsi[MAX_CH][4];	// [ch][scfsi_band]		1	bslbf
    struct _granule {
	struct _channel {
	    unsigned int  part2_3_length;	// [gr][ch]				12	uimsbf
	    unsigned int  big_values;	        // [gr][ch]				9	uimsbf
	    unsigned char global_gain;	        // [gr][ch]				8	uimsbf
	    unsigned char scalefac_compress;	// [gr][ch]				4	bslbf
	    unsigned char window_switching_flag;// [gr][ch]				1	bslbf
	    unsigned char block_type;		// [gr][ch]				2	bslbf
	    unsigned char mixed_block_flag;	// [gr][ch]				1	uimsbf
	    unsigned char table_select[2];	// [gr][ch][region]		5	bslbf
	    unsigned char subblock_gain[3];	// [gr][ch][window]		3	uimsbf
	    unsigned char region0_count;	// [gr][ch]				4	bslbf
	    unsigned char region1_count;	// [gr][ch]				3	bslbf
	    unsigned char preflag;		// [gr][ch]				1	bslbf
	    unsigned char scalefac_scale;	// [gr][ch]				1	bslbf
	    unsigned char count1table_select;	// [gr][ch]				1	bslbf
	    unsigned int cbmax; // [gr][ch] max non-zero sb
	    unsigned int count1;
	} ch[MAX_CH];
    } gr[2];

    unsigned char	scalefac_l[MAX_CH][23];			// [ch][sfb]			0..4	uimsbf
    unsigned char	scalefac_s[MAX_CH][13][3];		// [ch][sfb][window]	0..4	uimsbf

    unsigned char	num_channels;
    unsigned int	frame_duration;
} CT_MPEGInfo;

typedef struct _channel CT_Channel;
typedef struct _granule CT_Granule;

CT_Decoder* mpeg_stream_init(int cbuffer_size);

CT_MPEGInfo* mpeg_stream_decode(CT_Decoder *decoder);

#endif
