// -------------------------------------------------------
//
// Author: Thomas Lenart    
// Email:  thomas.lenart@es.lth.se
// Date:   March 2003
//
// -------------------------------------------------------
#include "main.h"

#ifdef DEBUG_REORDER
#include <stdio.h>
#endif

#include "memhandler.h"
#include "reorder.h"

static
unsigned char const sfb_48000_short[] = {
   4,  4,  4,  4,  6, 6, 10, 12, 14, 16, 20, 26, 66
};

static
unsigned char const sfb_44100_short[] = {
   4,  4,  4,  4,  6,  8,  10, 12, 14, 18, 22, 30, 56
};

static
unsigned char const sfb_32000_short[] = {
   4,  4,  4,  4,  6, 8, 12, 16, 20, 26, 34, 42, 12
};

static							// can be indexed with the value of frame_info->header.sampling_frequency
struct {
  unsigned char const *s;
} const sfbwidth_table[3] = {
  { sfb_44100_short },
  { sfb_48000_short },
  { sfb_32000_short }
};

void frame_reorder(CT_MPEGInfo *frame_info, CT_Channel *channel_info, fixed_point xr[MAX_SB*MAX_SS])
{
	fixed_point			ro[MAX_SB*MAX_SS];
	unsigned int		sb, width, sfb, source_index, dest_index, sfb_start, index_0, index_1, index_2, i;
	unsigned char const	*sfb_width;

	#ifdef DEBUG_REORDER
	FILE* file;
	#if DATA_WIDTH == 32
	file = fopen("debug_data/debug_reorder_32.txt", OUTPUT_FILE_TYPE);
	#else
	file = fopen("debug_data/debug_reorder_16.txt", OUTPUT_FILE_TYPE);
	#endif
	#endif

	if (channel_info->window_switching_flag && (channel_info->block_type == 2)) {

		sfb_width = sfbwidth_table[frame_info->header.sampling_frequency].s;

		if (channel_info->mixed_block_flag) {
			sfb_start = 3;
			sb = 2;
			dest_index = MAX_SS*2;
			for (i = 0; i < MAX_SS*2; i++)
				ro[i] = xr[i];
		}
		else {
			sfb_start = 0;
			sb = 0;
			dest_index = 0;
		}

		source_index = sb;
		for (sfb = sfb_start; sfb < 13; sfb++) {
			width = sfb_width[sfb];
			index_0 = source_index;				
			index_1 = source_index + width;				
			index_2 = source_index + (width << 1);				
			for (i = 0; i < width; i++) {
				ro[dest_index++] = xr[index_0++];
				ro[dest_index++] = xr[index_1++];
				ro[dest_index++] = xr[index_2++];
			}
			source_index += 3*width;
		}
		memcpy(&xr[sb*MAX_SS], &ro[sb*MAX_SS], (576 - MAX_SS * sb) * sizeof(fixed_point));
	}

	#ifdef DEBUG_REORDER
	for (i = 0; i < 576; i++)
		fprintf(file, "%d\n", xr[i]);
//		fprintf(file, "xr[x][%d][%d] %.10f\n", i / 18, i % 18, fixed_to_double(xr[i]));
	fclose(file);
	#endif
}
